/*-
 * Copyright (C)2003..2007 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: parser.h,v 1.16 2008/09/18 05:47:18 babolo Exp $
 */

#ifndef  _BABOLO_PARSER_H_
# define _BABOLO_PARSER_H_

__BEGIN_DECLS

/* For babolo_parm.flag                                                          */
# define Bpars_NPRM   0x0000FF /*                */
# define Bpars_LNSH 8
# define Bpars_VTLN/* 0x000100  VT -        */ (0x1 << Bpars_LNSH)
# define Bpars_CRLN/* 0x000200  CR -        */ (0x2 << Bpars_LNSH)
# define Bpars_LFLN/* 0x000400  LF -        */ (0x4 << Bpars_LNSH)
# define Bpars_XXLN/* 0x000700     */ (Bpars_VTLN | Bpars_LFLN | Bpars_CRLN)
# define Bpars_FAST   0x010000 /*                 */
# define Bpars_FIXB   0x020000 /*             */
# define Bpars_NOEN   0x040000 /*                */
# define Bpars_NOAL   0x080000 /*                        */
# define Bpars_FIRM   0x100000 /*                  */
# define Bpars_NONU   0x200000 /*                  */

/* For babolo_lexor.flag                                                         */
# define Bpars_AABS   0x200000 /*    lexgraf                */
# define Bpars_CEND   0x400000 /*                       */
# define Bpars_CMID   0x800000 /*        */

/* For babolo_opt.flags                                                          */
/*       Bpars_NOAL   0x080000                      */
# define Bpars_EOOP   0x100000 /*                                      */
/*       Bpars_NONU   0x200000      argv[0]                           */
# define Bpars_ESYM   0x400000 /*                          */
# define Bpars_EPRM   0x800000 /*     NULL      */

# define kukBpars_NPRM(x)   ((u_int32_t)(x) & Bpars_NPRM)
# define setBpars_NPRM(a,x) {(a) &= ~Bpars_NPRM; (a) |= kukBpars_NPRM(x);}
# define getBpars_NPRM(a)   ((a) & Bpars_NPRM)

typedef struct babolo_parm {
    u_int32_t flag;
    int       argc;
    size_t    cntr;
    u_char  *(argv[0]);
} babolo_parm;

typedef struct babolo_opt {
    struct babolo_opt *next;/*  (sic!)               */
    u_char     *arg;        /*   ( NULL)            */
    u_char     *plc;        /*             */
    u_char    **argv;       /* -                          */
    u_char     *opts;       /*            */
    u_int32_t   flags;      /*                                     */
    u_int32_t   argc;       /*    argv                   */
    u_int32_t   ind;        /*            */
} babolo_opt;

typedef struct babolo_lexor {
    u_char   *retable;      /*                      */
    void     *printable;    /*                                 */
    u_int32_t flag;         /*                                     */
    u_int32_t szt;          /*      szp + 1  */
    u_int32_t szz;          /*     szp + 1 */
    u_char    szp;          /*                      */
#if defined(__GNUC__) && __GNUC__ > 2
    u_char    lexgraf[];    /*                                    */
#else
    u_char    lexgraf[0];   /*                                    */
#endif
       /*****************************************************************
        *      :                            *
        *    N(1 ) + 1 -   (, )             *
        *      (1 ), (szp )                         *
        *      ...                                                      *
        *    (szp )  ,              *
        *                                                               *
        *      szp + 1,  Bpars_AABS ,       *
        *       + szp           *
        *    ,             *
        *   .                                                    *
        *      szz,             *
        *    ,               *
        *    -    .                *
        *****************************************************************/
} babolo_lexor;

babolo_parm *babolo_getparm __P((u_int32_t flag, u_char **instring, const u_char *strend, u_char absq));
void        babolo_freeparm __P((babolo_parm *parsed));
int          babolo_openopt __P((u_int32_t flags, babolo_opt **opt, int argc, char *const *argv, const char *ostr));
int           babolo_getopt __P((babolo_opt **opt));
char      *babolo_getoptarg __P((babolo_opt **opt, int argn));
char         *babolo_getarg __P((babolo_opt **opt));
int         babolo_closeopt __P((babolo_opt **opt));
void         babolo_dumpopt __P((babolo_opt *opt));
u_int32_t   babolo_testchar __P((const babolo_lexor *lext, u_int32_t state, const u_char **next));
u_int32_t   babolo_testword __P((const babolo_lexor *lext, const u_char *word));
u_int32_t     babolo_goword __P((const babolo_lexor *lext, const u_char **word));
void        babolo_treedump __P((const babolo_lexor *lext));

__END_DECLS

#endif /* _BABOLO_PARSER_H_ */
